/*
 *  @title    Control of a simple boost converter on PWM channel #0
 *	@author   imperix SA (dev@imperix.ch)
 *	@file     user.c
 */

#include "Driver/peripherals.h"							// Available DSP peripherals
#include "Core/interrupts.h"							// Interrupt-related routines
#include "user.h"										// This file's header.
#include <stdio.h>

#include "../API/transformations.h"						// Coordinate transformations
#include "../API/controllers.h"							// Discrete-time controllers
#include "../API/PLLs.h"								// PLLs


/**
 * Global variables
 * Please pay attention not to define variables that are unnecessarily global !
 */
float Upv, Epv;											// PV voltage (actual and converter voltage)
float Ipv, Ipv_ref;										// PV current (actual and setpoint)
float Udc;												// DC bus voltage

float Ipv_prev, Ipv_delta;								// Last value and delta of the PV current (MPPT)
float Ppv, Ppv_prev, Ppv_delta;							// Last value and delta of the PV power (MPPT)

PIDController Ipv_reg;									// Controller for PV current control
Uint16 enable_MPPT;										// Boolean to activate/deactivate the MPP-Tracking


/**
 * The user should put here his/her initialization routines.
 * This function is called only once at startup.
 */
USER_SAFE UserInit(void)
{
	// Configure a 10 kHz switching frequency on clock source #3:
	SetFreqGenPeriod(3, (int)(SWITCHING_PERIOD/FPGA_CLK_PERIOD));

	// Map PWM channel #0 to clksource #3 with a triangular carrier and a 400 ns deadtime:
	ConfigPWMChannel(0, 3, SAWTOOTH, (int)(400e-9/FPGA_CLK_PERIOD));
	ActivatePWMChannel(0);								// Activate the PWM #0 channel

	// Channel #0 is configured with a phase of zero with respect to the frequency generator:
	SetPWMPhase(0, 0.0);

	// Analog input #0 to #2 with corrective gain and offset:
	SetADCAdjustments(0, 6.10e-3/4, 0.0);				// Nominal sensitivity and x4 gain (Ipv)
	SetADCAdjustments(1, 57.4e-3/8, 0.0);				// Nominal sensitivity and x8 gain (Upv)
	SetADCAdjustments(2, 113e-3/8, 0.0);				// Nominal sensitivity and x8 gain (Udc)

	// Configure UsrInt1 on the PWM channel #0:
	RegisterExt1Interrupt(&UserInterrupt1, 3, 0.0, 0);
	
	// Configure UsrInt2 on the CPU timer with a 10ms interrupt period:
	RegisterTimerInterrupt(&UserInterrupt2, 10000); 
	
	// Configure the sampling in the middle of the period of the FreqGen #3:
	ConfigSampling(3, 0.5);

	// Configure the PI current controller:
	#define Kp 8.0
	#define Ki 0.01
	#define Kd 0.0
	ConfigPIDController(&Ipv_reg, Kp, Ki, Kd, 15, -15, SAMPLING_PERIOD, 10);

	// Initialize the setpoints and other global variables:
	Ipv_ref = 0.0;
	enable_MPPT = 0;

	return SAFE;										// Return 0 if no error
}


/**
 * First user-level interrupt service routine.
 * This routine is synchronous to the PWM #0 period
 */
USER_SAFE UserInterrupt1(void)
{
	// Measure all the necessary quantities:
	Upv = GetADC(1);									// Voltage on the PV panel
	Ipv = -GetADC(0);									// PV panel current (sensor is positive outbound)
	Udc = GetADC(2);									// DC bus voltage

	// Execute the current controller:
	Epv = Upv - RunPIController(&Ipv_reg, Ipv_ref - Ipv);

	// Refresh the PWM duty-cycle of PWM channel #0:
	SetPWMDutyCycle(0, Epv/Udc);

	// Send the updated PWM parameters to the modulator (in the FPGA):
	UpdatePWMData();

	return SAFE;										// Return a SAFE status (or UNSAFE otherwise)
}


/**
 * First user-level interrupt service routine.
 * This routine is used with a relatively long interrupt period and is only
 * responsible for executing the MPPT "pertub and assess" algorithm
 */
USER_SAFE UserInterrupt2(void)
{
	// Compute the power drawn from the PV panel:
	Ppv = Upv*Ipv;

	// Compute the differences since the last interrupt execution:
	Ppv_delta = Ppv - Ppv_prev;
	Ipv_delta = Ipv - Ipv_prev;

	// Execute the MPP-tracking algorithm:
	if (enable_MPPT){
		if (Ppv_delta >= 0){
			// If there is any improvement, keep going...
			if (Ipv_delta >=0 ){ Ipv_ref += 0.01; } else{ Ipv_ref -= 0.01; }
		}
		else{
			// Else, do the opposite...
			if (Ipv_delta >= 0){ Ipv_ref -= 0.01; } else{ Ipv_ref += 0.01; }
		}
		// Force the MPPT boundaries:
		if (Ipv_ref<=0.5) Ipv_ref = 0.0;
		if (Ipv_ref>=15) Ipv_ref = 10.0;
	}
	else {
														// Else: Ipv_ref is left unset
														// It can be imposed directly by the used (e.g. through the console)
	}

	// Update the previous values:
	Ppv_prev = Ppv;
	Ipv_prev = Ipv;

	// Configure the user LED on the frontpanel to indicate the activation status of the MPPT:
	if (enable_MPPT==1){ SetUserLED(LED_GREEN); }
	else{ SetUserLED(LED_OFF); }

	return SAFE;
}


/*
 * This routine aims to handle critical errors by requesting an appropriate answer at the user level.
 * Typically, the application should be blocked and the core will wait for manual clearing of the error.
 */
void UserError(ERROR_SOURCE source)
{
	printf("\nUserError() has been triggered!");		// Only a warning message is displayed here...
}
